@@ -12,9 +12,16 @@
 #include <stdio.h>
 #include <unistd.h>
 #include <sys/types.h>
+#include <ctype.h>
+#include <stdarg.h>
+#include <string.h>
+
+#ifdef USE_CURSES
 #include <curses.h>
-#include "yahoolib.h"
-#include "yahoolib-proto.h"
+#endif
+
+#include "libyahoo.h"
+#include "libyahoo-proto.h"
 
 struct statusrec {
   char *id;
@@ -35,23 +42,35 @@ char *cur_target = NULL;
 char *prev_target = NULL;
 char *reply_target = NULL;
 int cur_status = YAHOO_STATUS_AVAILABLE;
+int cur_target_changed = 1;
+int cur_status_changed = 1;
 
+#ifdef USE_CURSES
 WINDOW *win;
 WINDOW *displaywin;
 WINDOW *sepwin;
 WINDOW *inputwin;
+#endif
 
 void main_loop(void);
 void process_input(void);
 void handle_io(void);
+int is_integer(char *);
+void myprintf(char *, ...);
+
+#ifdef USE_CURSES
 void separatewin(char *, char *);
 void init_curses(void);
+#else
+void show_prompt(void);
+#endif
 
 int main(int argc, char *argv[]) {
   int i;
   struct yahoo_buddy *tmpbuddy;
   struct statusrec *tmprec;
   char *id;
+  char *tmp;
 
   if(argc > 3) {
     fprintf(stderr, "Usage:  %s [<username>] [<password>]\n", argv[0]);
@@ -66,13 +85,17 @@ int main(int argc, char *argv[]) {
   } else if(argc == 2) {
     username = (char *) malloc(strlen(argv[1]) + 1);
     strncpy(username, argv[1], strlen(argv[1]));
-    password = getpass("Password: ");
+    tmp = getpass("Password: ");
+    password = (char *) malloc(strlen(tmp) + 1);
+    strncpy(password, tmp, strlen(tmp));
   } else if(argc == 1) {
     username = (char *) malloc(100);
     printf("Yahoo ID: ");
     fflush(stdout);
     scanf("%s", username);
-    password = getpass("Password: ");
+    tmp = getpass("Password: ");
+    password = (char *) malloc(strlen(tmp) + 1);
+    strncpy(password, tmp, strlen(tmp));
   }
 
   options.connect_mode = YAHOO_CONNECT_NORMAL;
@@ -112,13 +135,15 @@ int main(int argc, char *argv[]) {
     i--;
   }
 
+#ifdef USE_CURSES
   init_curses();
+#endif
 
-  wprintw(displaywin, "Thank you for using Ari's Yahoo Client.\n");
-  wprintw(displaywin, "Ari's Yahoo Client is copyright (C) 2000 Ari Johnson\n");
-  wprintw(displaywin, "and is covered by the GNU General Public License.\n\n");
-  wprintw(displaywin, "Yahoo Messenger interface provided by yahoolib,\n");
-  wprintw(displaywin, "as distributed with GTKYahoo.\n\n");
+  myprintf("Thank you for using Ari's Yahoo Client.\n");
+  myprintf("Ari's Yahoo Client is copyright (C) 2000 Ari Johnson\n");
+  myprintf("and is covered by the GNU General Public License.\n\n");
+  myprintf("Yahoo Messenger interface provided by yahoolib,\n");
+  myprintf("as distributed with GTKYahoo.\n\n");
 
   main_loop();
 
@@ -214,7 +239,7 @@ void show_userlist(void) {
     in_chat = get_in_chat(id);
 
     if(!in_pager && !in_chat)
-      wprintw(displaywin, "   %s\n", id);
+      myprintf("   %s\n", id);
     i++;
   }
 
@@ -227,8 +252,8 @@ void show_userlist(void) {
     in_pager = get_in_pager(id);
     in_chat = get_in_chat(id);
 
-    if(in_pager || in_chat && status != YAHOO_STATUS_AVAILABLE)
-      wprintw(displaywin, "*  %s [%s]\n", id, yahoo_get_status_string(status));
+    if((in_pager || in_chat) && status != YAHOO_STATUS_AVAILABLE)
+      myprintf("*  %s [%s]\n", id, yahoo_get_status_string(status));
     i++;
   }
 
@@ -241,8 +266,8 @@ void show_userlist(void) {
     in_pager = get_in_pager(id);
     in_chat = get_in_chat(id);
 
-    if(in_pager || in_chat && status == YAHOO_STATUS_AVAILABLE)
-      wprintw(displaywin, "*  %s\n", id);
+    if((in_pager || in_chat) && status == YAHOO_STATUS_AVAILABLE)
+      myprintf("*  %s\n", id);
     i++;
   }
 }
@@ -290,6 +315,7 @@ int do_remove(char *id) {
   return 1;
 }
 
+#ifdef USE_CURSES
 void init_curses(void) {
   int maxx, maxy;
 
@@ -305,6 +331,7 @@ void init_curses(void) {
 
   scrollok(inputwin, TRUE);
   keypad(inputwin, TRUE);
+  nodelay(inputwin, TRUE);
   scrollok(displaywin, TRUE);
 
   separatewin(NULL, NULL);
@@ -345,13 +372,14 @@ void separatewin(char *s, char *t) {
     buff[i++] = '>';
   }
 
-  for(; i < (maxx - 5); i++)
+  for(; i < (maxx - 6); i++)
     buff[i] = '-';
 
-  if(i <= (maxx - 5)) {
+  if(i <= (maxx - 6)) {
     buff[i++] = '1';
     buff[i++] = '.';
-    buff[i++] = '7';
+    buff[i++] = '1';
+    buff[i++] = '0';
     buff[i++] = '-';
     buff[i++] = '-';
   }
@@ -361,14 +389,27 @@ void separatewin(char *s, char *t) {
   wmove(sepwin, 0, 0);
   waddstr(sepwin, buff);
   wrefresh(sepwin);
+  wrefresh(inputwin);
 }
+#else
+void show_prompt(void) {
+  if(cur_status != YAHOO_STATUS_AVAILABLE)
+    printf("[%s] ", yahoo_get_status_string(cur_status));
+  printf("%s:  ", cur_target ? cur_target : "<none>");
+  fflush(stdout);
+}
+#endif
 
 void main_loop(void) {
   fd_set inset;
 
   quit_flag = 0;
 
-  wprintw(displaywin, "Connected.\n");
+  myprintf("Connected.\n");
+
+#ifndef USE_CURSES
+  show_prompt();
+#endif
 
   while(!quit_flag) {
     FD_ZERO(&inset);
@@ -380,23 +421,36 @@ void main_loop(void) {
       return;
     }
 
-    if(FD_ISSET(fileno(stdin), &inset))
+    if(FD_ISSET(fileno(stdin), &inset)) {
       process_input();
-    if(FD_ISSET(context->sockfd, &inset))
+#ifdef USE_CURSES
+      wrefresh(inputwin);
+#endif
+    }
+    if(FD_ISSET(context->sockfd, &inset)) {
       handle_io();
+#ifdef USE_CURSES
+      wrefresh(displaywin);
+      wrefresh(inputwin);
+#endif
+    }
 
-    wrefresh(displaywin);
-
-    separatewin((cur_status != YAHOO_STATUS_AVAILABLE)
-		? yahoo_get_status_string(cur_status) : NULL,
-		cur_target);
-
-    wrefresh(inputwin);
+    if(cur_status_changed || cur_target_changed) {
+#ifdef USE_CURSES
+      separatewin((cur_status != YAHOO_STATUS_AVAILABLE)
+			? yahoo_get_status_string(cur_status) : NULL,
+			cur_target);
+#endif
+      cur_status_changed = 0;
+      cur_target_changed = 0;
+    }
   }
 
   yahoo_cmd_logoff(context);
 
+#ifdef USE_CURSES
   endwin();
+#endif
 }
 
 void process_command(char *orig) {
@@ -414,32 +468,37 @@ void process_command(char *orig) {
         *args++ = '\0';
 
     if(!strcasecmp(cmd, "help")) {
-      wprintw(displaywin, "Ari's Yahoo Client Help\n");
-      wprintw(displaywin, "--------------------------------------\n");
-      wprintw(displaywin, "Commands:\n");
-      wprintw(displaywin, "  /help            Display this screen.\n");
-      wprintw(displaywin, "  /clear           Clear the screen.\n");
-      wprintw(displaywin, "  /c               Alias for /clear.\n");
-      wprintw(displaywin, "  /who             Display friends list.\n");
-      wprintw(displaywin, "  /w               Alias for /who.\n");
-      wprintw(displaywin, "  /msg <ID> <msg>  Send <msg> to <ID>.\n");
-      wprintw(displaywin, "  /m <ID> <msg>    Alias for /msg.\n");
-      wprintw(displaywin, "  /query <ID>      Select <ID> as target.\n");
-      wprintw(displaywin, "  /query           Unselect target.\n");
-      wprintw(displaywin, "  /q [<ID>]        Alias for /query.\n");
-      wprintw(displaywin, "  /prev            Select previous target.\n");
-      wprintw(displaywin, "  /p               Alias for /prev.\n");
-      wprintw(displaywin, "  /reply <msg>     Reply to last message.\n");
-      wprintw(displaywin, "  /r <msg>         Alias for /reply.\n");
-      wprintw(displaywin, "  /status <#>      Set status to #.\n");
-      wprintw(displaywin, "  /add <ID>        Add a friend.\n");
-      wprintw(displaywin, "  /remove <ID>     Remove a friend.\n");
-      wprintw(displaywin, "  /quit            Exit the program.\n");
-      wprintw(displaywin, "--------------------------------------\n");
+      myprintf("Ari's Yahoo Client Help\n");
+      myprintf("--------------------------------------\n");
+      myprintf("Commands:\n");
+      myprintf("  /help            Display this screen.\n");
+#ifdef USE_CURSES
+      myprintf("  /clear           Clear the screen.\n");
+      myprintf("  /c               Alias for /clear.\n");
+#endif
+      myprintf("  /who             Display friends list.\n");
+      myprintf("  /w               Alias for /who.\n");
+      myprintf("  /msg <ID> <msg>  Send <msg> to <ID>.\n");
+      myprintf("  /m <ID> <msg>    Alias for /msg.\n");
+      myprintf("  /query <ID>      Select <ID> as target.\n");
+      myprintf("  /query           Unselect target.\n");
+      myprintf("  /q [<ID>]        Alias for /query.\n");
+      myprintf("  /prev            Select previous target.\n");
+      myprintf("  /p               Alias for /prev.\n");
+      myprintf("  /reply <msg>     Reply to last message.\n");
+      myprintf("  /r <msg>         Alias for /reply.\n");
+      myprintf("  /statuslist      Lists possible statuses.\n");
+      myprintf("  /status <s>      Set status to <s>.\n");
+      myprintf("  /add <ID>        Add a friend.\n");
+      myprintf("  /remove <ID>     Remove a friend.\n");
+      myprintf("  /quit            Exit the program.\n");
+      myprintf("--------------------------------------\n");
+#ifdef USE_CURSES
     } else if(!strcasecmp(cmd, "clear") || !strcasecmp(cmd, "c")) {
       wclear(displaywin);
       wmove(displaywin, 0, 0);
       wrefresh(displaywin);
+#endif
     } else if(!strcasecmp(cmd, "who") || !strcasecmp(cmd, "w")) {
       show_userlist();
     } else if(!strcasecmp(cmd, "msg") || !strcasecmp(cmd, "m")) {
@@ -447,20 +506,20 @@ void process_command(char *orig) {
       char *message;
       target = args;
       if(!target) {
-        wprintw(displaywin, "You must supply a target and a message.\n");
+        myprintf("You must supply a target and a message.\n");
         return;
       }
       while(*target == ' ')
         *target++ = '\0';
       message = strchr(target, ' ');
       if(!message) {
-        wprintw(displaywin, "You must supply a target and a message.\n");
+        myprintf("You must supply a target and a message.\n");
         return;
       }
       while(*message == ' ')
         *message++ = '\0';
       yahoo_cmd_msg(context, username, target, message);
-      wprintw(displaywin, "-> %s: %s\n", target, message);
+      myprintf("-> %s: %s\n", target, message);
     } else if(!strcasecmp(cmd, "query") || !strcasecmp(cmd, "q")) {
       if(cur_target) {
         if(prev_target)
@@ -470,126 +529,155 @@ void process_command(char *orig) {
         prev_target[strlen(cur_target)] = '\0';
         free(cur_target);
         cur_target = NULL;
+        cur_target_changed = 1;
       }
       if(args && *args) {
         cur_target = (char *) malloc(strlen(args) + 1);
         strncpy(cur_target, args, strlen(args));
         cur_target[strlen(args)] = '\0';
+        cur_target_changed = 1;
       }
     } else if(!strcasecmp(cmd, "prev") || !strcasecmp(cmd, "p")) {
       if(!prev_target) {
-        wprintw(displaywin, "No previous target.\n");
+        myprintf("No previous target.\n");
       } else {
         char *tmp;
         tmp = prev_target;
         prev_target = cur_target;
         cur_target = tmp;
+        cur_target_changed = 1;
       }
     } else if(!strcasecmp(cmd, "reply") || !strcasecmp(cmd, "r")) {
       if(!reply_target) {
-        wprintw(displaywin, "No target to which to reply.\n");
+        myprintf("No target to which to reply.\n");
       } else if(!args || !*args) {
-        wprintw(displaywin, "No message.\n");
+        myprintf("No message.\n");
       } else {
         yahoo_cmd_msg(context, username, reply_target, args);
-        wprintw(displaywin, "-> %s: %s\n", reply_target, args);
+        myprintf("-> %s: %s\n", reply_target, args);
       }
+    } else if(!strcasecmp(cmd, "statuslist")) {
+      int i;
+      for(i = 0; i <= YAHOO_STATUS_IDLE; i++)
+        if(i != YAHOO_STATUS_CUSTOM && yahoo_get_status_string(i))
+          myprintf("%8d: %s\n", i, yahoo_get_status_string(i));
+      myprintf("<string>: Custom message\n");
     } else if(!strcasecmp(cmd, "status")) {
       if(args && *args) {
         int newstatus;
-        newstatus = atoi(args);
-        if(yahoo_get_status_string(newstatus)) {
-          if(newstatus == YAHOO_STATUS_AVAILABLE)
-            yahoo_cmd_set_back_mode(context, newstatus);
-          else
-            yahoo_cmd_set_away_mode(context, newstatus);
-          cur_status = newstatus;
+        if(!is_integer(args)) {
+          yahoo_cmd_set_away_mode(context, YAHOO_STATUS_CUSTOM, args);
+          cur_status = YAHOO_STATUS_CUSTOM;
+          cur_status_changed = 1;
         } else {
-          int i;
-          wprintw(displaywin, "Invalid status number.\n");
-          for(i = 0; i <= YAHOO_STATUS_IDLE; i++)
-            if(yahoo_get_status_string(i))
-              wprintw(displaywin, "%5d: %s\n", i, yahoo_get_status_string(i));
+          newstatus = atoi(args);
+          if(yahoo_get_status_string(newstatus)) {
+            if(newstatus == YAHOO_STATUS_AVAILABLE)
+              yahoo_cmd_set_back_mode(context, newstatus, NULL);
+            else
+              yahoo_cmd_set_away_mode(context, newstatus, NULL);
+            cur_status = newstatus;
+            cur_status_changed = 1;
+          } else {
+            myprintf("Invalid status number.\n");
+          }
         }
       } else {
-        int i;
-        wprintw(displaywin, "You must specify a status number.\n");
-        for(i = 0; i <= YAHOO_STATUS_IDLE; i++)
-          if(yahoo_get_status_string(i))
-            wprintw(displaywin, "%5d: %s\n", i, yahoo_get_status_string(i));
+        cur_status = YAHOO_STATUS_AVAILABLE;
+        yahoo_cmd_set_back_mode(context, cur_status, NULL);
+        cur_status_changed = 1;
       }
     } else if(!strcasecmp(cmd, "add")) {
       if(args && *args) {
         if(do_add(args))
-          wprintw(displaywin, "Could not add friend.\n");
+          myprintf("Could not add friend.\n");
         else
-          wprintw(displaywin, "Added `%s' to friends list.\n", args);
+          myprintf("Added `%s' to friends list.\n", args);
       } else {
-        wprintw(displaywin, "You must give a Yahoo ID to add to your list.\n");
+        myprintf("You must give a Yahoo ID to add to your list.\n");
       }
     } else if(!strcasecmp(cmd, "remove")) {
       if(args && *args) {
         if(do_remove(args))
-          wprintw(displaywin, "Could not remove friend.\n");
+          myprintf("Could not remove friend.\n");
         else
-          wprintw(displaywin, "Removed `%s' from friends list.\n", args);
+          myprintf("Removed `%s' from friends list.\n", args);
       } else {
-        wprintw(displaywin,
-		"You must give a Yahoo ID to remove from your list.\n");
+        myprintf("You must give a Yahoo ID to remove from your list.\n");
       }
     } else if(!strcasecmp(cmd, "quit")) {
       quit_flag = 1;
     } else {
-      wprintw(displaywin, "Invalid command.\n");
+      myprintf("Invalid command.\n");
     }
   } else {
     if(!cur_target) {
-      wprintw(displaywin, "No target selected.\n");
+      myprintf("No target selected.\n");
       return;
     }
     yahoo_cmd_msg(context, username, cur_target, orig);
-    wprintw(displaywin, "-> %s: %s\n", cur_target, orig);
+    myprintf("-> %s: %s\n", cur_target, orig);
   }
 }
 
+#ifdef USE_CURSES
 void process_input(void) {
   static char buff[1024];
   static char *ptr = buff;
   int c;
 
   c = wgetch(inputwin);
+  while(c != ERR) {
+    switch(c) {
+      case '\n':
+      case '\r':
+        *ptr = '\0';
+        ptr = buff;
+        waddch(inputwin, '\n');
+        wrefresh(inputwin);
+        process_command(buff);
+        wrefresh(displaywin);
+        break;
+      case '\b':
+      case 127:
+        if(ptr > buff) {
+          ptr--;
+          waddch(inputwin, '\b');
+          waddch(inputwin, ' ');
+          waddch(inputwin, '\b');
+        }
+        break;
+      default:
+        *ptr++ = c;
+        waddch(inputwin, c);
+    }
 
-  switch(c) {
-    case '\n':
-    case '\r':
-      *ptr = '\0';
-      ptr = buff;
-      waddch(inputwin, '\n');
-      wrefresh(inputwin);
-      process_command(buff);
-      wrefresh(displaywin);
-      break;
-    case '\b':
-    case 127:
-      if(ptr > buff) {
-        ptr--;
-        waddch(inputwin, '\b');
-        waddch(inputwin, ' ');
-        waddch(inputwin, '\b');
-      }
-      break;
-    default:
-      *ptr++ = c;
-      waddch(inputwin, c);
+    c = wgetch(inputwin);
   }
 }
+#else
+void process_input(void) {
+  char buff[1024];
+  char *ptr;
+
+  fgets(buff, 1024, stdin);
+  ptr = strchr(buff, '\n');
+  if(ptr)
+    *ptr = '\0';
+  ptr = strchr(buff, '\r');
+  if(ptr)
+    *ptr = '\0';
+  process_command(buff);
+  show_prompt();
+}
+#endif
 
 void process_packet_status(struct yahoo_packet *pkt) {
   int i;
 
   if(pkt->service == YAHOO_SERVICE_LOGOFF &&
      !strcmp(pkt->active_id, username)) {
-    wprintw(displaywin, "Disconnected.\n");
+    myprintf("Disconnected.\n");
     quit_flag = 1;
   }
 
@@ -608,12 +696,16 @@ void process_packet_status(struct yahoo_packet *pkt) {
     }
 
     if(pkt->service == YAHOO_SERVICE_LOGON)
-      wprintw(displaywin, "%s has connected.\n", id);
+      myprintf("%s has connected.\n", id);
     else if(pkt->service == YAHOO_SERVICE_LOGOFF)
-      wprintw(displaywin, "%s has disconnected.\n", id);
+      myprintf("%s has disconnected.\n", id);
     else
-      wprintw(displaywin, "%s is now %s.\n", id,
+      myprintf("%s is now %s.\n", id,
 		yahoo_get_status_string(get_status(id)));
+
+#ifndef USE_CURSES
+    show_prompt();
+#endif
   }
 }
 
@@ -623,10 +715,10 @@ void process_packet_message(struct yahoo_packet *pkt) {
 
   if(pkt->msg) {
     if(pkt->msgtype == YAHOO_MSGTYPE_BOUNCE) {
-      wprintw(displaywin, "%s -> %s: <Message not sent, user not online.>\n",
+      myprintf("%s -> %s: <Message not sent, user not online.>\n",
 		pkt->active_id, pkt->msg_id);
     } else {
-      wprintw(displaywin, "%s -> %s: %s\n", pkt->msg_id, pkt->active_id,
+      myprintf("%s -> %s: %s\n", pkt->msg_id, pkt->active_id,
 		pkt->msg);
       if(reply_target)
         free(reply_target);
@@ -634,11 +726,19 @@ void process_packet_message(struct yahoo_packet *pkt) {
       strncpy(reply_target, pkt->msg_id, strlen(pkt->msg_id));
       reply_target[strlen(pkt->msg_id)] = '\0';
     }
+
+#ifndef USE_CURSES
+    show_prompt();
+#endif
   }
 }
 
 void process_packet_ping(struct yahoo_packet *pkt) {
-  wprintw(displaywin, "Ping.\n");
+  myprintf("Ping.\n");
+
+#ifndef USE_CURSES
+  show_prompt();
+#endif
 }
 
 void handle_io(void) {
@@ -654,16 +754,15 @@ void handle_io(void) {
     pkt = yahoo_parsepacket(context, rawpkt);
 
     if(debug_packets) {
-      wprintw(displaywin, "Received packet:\n");
-      wprintw(displaywin, "\tService = %s\n",
-		yahoo_get_service_string(pkt->service));
-      wprintw(displaywin, "\tReal ID = %s\n", pkt->real_id);
-      wprintw(displaywin, "\tActive ID = %s\n", pkt->active_id);
-      wprintw(displaywin, "\tConnection ID = %X\n", pkt->connection_id);
-      wprintw(displaywin, "\tMagic ID = %X\n", pkt->magic_id);
-      wprintw(displaywin, "\tUnknown Flag 1 = %X\n", pkt->unknown1);
-      wprintw(displaywin, "\tMessage Type = %X\n", pkt->msgtype);
-      wprintw(displaywin, "\tRaw Content = %s\n", rawpkt->content);
+      myprintf("Received packet:\n");
+      myprintf("\tService = %s\n", yahoo_get_service_string(pkt->service));
+      myprintf("\tReal ID = %s\n", pkt->real_id);
+      myprintf("\tActive ID = %s\n", pkt->active_id);
+      myprintf("\tConnection ID = %X\n", pkt->connection_id);
+      myprintf("\tMagic ID = %X\n", pkt->magic_id);
+      myprintf("\tUnknown Flag 1 = %X\n", pkt->unknown1);
+      myprintf("\tMessage Type = %X\n", pkt->msgtype);
+      myprintf("\tRaw Content = %s\n", rawpkt->content);
     }
 
     switch(pkt->service) {
@@ -698,3 +797,38 @@ void handle_io(void) {
   }
 }
 
+int is_integer(char *s) {
+  char *ptr;
+
+  if(!s || !*s)
+    return 0;
+
+  ptr = s;
+  if(((*ptr == '-') || (*ptr == '+')) && ptr[1])
+    ptr++;
+  while(isdigit(*ptr))
+    ptr++;
+  if(*ptr)
+    return 0;
+
+  return 1;
+}
+
+#ifdef USE_CURSES
+void myprintf(char *fmt, ...) {
+  va_list ap;
+
+  va_start(ap, fmt);
+  vwprintw(displaywin, fmt, ap);
+  va_end(ap);
+}
+#else
+void myprintf(char *fmt, ...) {
+  va_list ap;
+
+  va_start(ap, fmt);
+  vprintf(fmt, ap);
+  va_end(ap);
+}
+#endif
+
